from typing import Dict

from wsknn.preprocessing.static_parsers.parse import parse_stream


def _get_header(header, sep, session_index: int, product_index: int, time_index: int, action_index: int = None):
    """
    Function gets header names.

    Parameters
    ----------
    header : str
        Line with a header.

    sep : str
        Separator.

    session_index : int

    product_index : int

    time_index : int

    action_index : int, optional

    Returns
    -------
    header_map : Dict
        Dictionary with pairs {index: name}
    """
    try:
        header_names = header.split(sep)
    except Exception as _:
        raise AttributeError('Cannot parse header from file, wrong separator!')

    # Create dictionary

    if action_index is not None:
        header_map = {
            session_index: header_names[session_index],
            action_index: header_names[action_index],
            product_index: header_names[product_index],
            time_index: header_names[time_index]
        }
    else:
        header_map = {
            session_index: header_names[session_index],
            product_index: header_names[product_index],
            time_index: header_names[time_index]
        }

    return header_map


def parse_flat_file_fn(dataset: str,
                       sep: str,
                       session_index: int,
                       product_index: int,
                       time_index: int,
                       action_index: int = None,
                       use_header_row: bool = False,
                       time_to_numeric=False,
                       time_to_datetime=False,
                       datetime_format='',
                       allowed_actions: Dict = None,
                       purchase_action_name=None,
                       ignore_errors: bool = True,
                       progress_bar: bool = False):
    """
    Function parses data from flat file into item-sessions and session-items maps.

    Parameters
    ----------
    dataset : str
        Input file.

    sep : str
        Separator used to separate values.

    session_index : int
        The index of the session.

    product_index : int
        The index of the product.

    time_index : int
        The index of the event timestamp.

    action_index : int, optional
        The index of the event action.

    use_header_row : bool, default = False
        Use first row values as a header.

    time_to_numeric : bool, default = False
        Transforms input timestamps to float values.

    time_to_datetime : bool, default = False
        Transforms input timestamps to datatime objects. Setting ``datetime_format`` parameter is required.

    datetime_format : str
        The format of datetime object.

    allowed_actions : Dict, optional
        Allowed actions and their weights.

    purchase_action_name: Any, optional
        The name of the final action (it is required to apply weight into the session vector).

    ignore_errors : bool, default=True
        Ignore rows that raise exceptions.

    progress_bar : bool
        Show parsing progress.

    Returns
    -------
    items, sessions : Items, Sessions
        The mappings of item-session and session-items.
    """

    with open(dataset, 'r') as fin:
        if use_header_row:
            header = _get_header(next(fin),
                                 sep=sep,
                                 session_index=session_index,
                                 product_index=product_index,
                                 time_index=time_index,
                                 action_index=action_index)
        else:
            header = None

        items, sessions = parse_stream(
            events=fin,
            sep=sep,
            allowed_actions=allowed_actions,
            purchase_action_name=purchase_action_name,
            session_index=session_index,
            product_index=product_index,
            time_index=time_index,
            action_index=action_index,
            time_to_numeric=time_to_numeric,
            time_to_datetime=time_to_datetime,
            datetime_format=datetime_format,
            ignore_errors=ignore_errors,
            header_names=header,
            progress_bar=progress_bar
        )

    return items, sessions



from wsknn import Items, Sessions
from wsknn.preprocessing.parse_static import parse_flat_file

CSVFILE = 'tdata/events1.csv'
SESSION_IDX = 3
PRODUCT_IDX = 4
ACTION_IDX = 0
TIME_IDX = 1
POSSIBLE_ACTIONS = {'products_view': 0.1,
                    'purchase': 1,
                    'add_to_wishlist': 0.3,
                    'add_to_cart': 0.5}
PURCHASE_ACTION_NAME = 'purchase'


def test_parse_flat_file():
    parsed_items, parsed_sessions = parse_flat_file(
        dataset=CSVFILE,
        sep=',',
        session_index=SESSION_IDX,
        product_index=PRODUCT_IDX,
        action_index=ACTION_IDX,
        time_index=TIME_IDX,
        use_header_row=True,
        time_to_numeric=True
    )

    assert isinstance(parsed_items, Items)
    assert isinstance(parsed_sessions, Sessions)



def get_smaller_value(val1, val2):
    """
    Function compares two values and returns smaller one.

    Parameters
    ----------
    val1 : number

    val2 : number

    Returns
    -------
    : number
    """
    if val1 < val2:
        return val1
    return val2


def get_larger_value(val1, val2):
    """
    Function compares two values and returns larger one.

    Parameters
    ----------
    val1 : number

    val2 : number

    Returns
    -------
    : number
    """
    if val1 > val2:
        return val1
    return val2



from typing import Dict, Iterable


def check_event_keys_and_values(event: Dict,
                                session_id_key: str,
                                product_key: str,
                                time_key: str,
                                action_key: str = None):
    """
    Function checks if event has its all keys.

    Parameters
    ----------
    event : Dict

    session_id_key : str

    product_key : str

    action_key : str

    time_key : str

    Returns
    -------
    : Dict
        Empty dict if there is a missing key.
    """

    keys = {session_id_key, product_key, time_key}

    if action_key is not None:
        keys.add(action_key)

    if not keys.issubset(set(event.keys())):
        return {}

    return event


def is_user_item_interaction(e_action: str, allowed_actions: Iterable) -> bool:
    """
    Checks if event has a valid action type.

    Parameters
    ----------
    e_action : str
        The name of the event action.

    allowed_actions : Iterable
        The set, list or array with allowed actions.

    Returns
    -------
    : bool
        ``True`` if ``e_action`` is in the list of allowed interactions.
    """

    if e_action in allowed_actions:
        return True
    return False



import pathlib
from typing import Dict, Tuple

from wsknn.preprocessing.static_parsers.csv_parser.parse import parse_csv_fn
from wsknn.preprocessing.static_parsers.flat_file_parser.parse import parse_flat_file_fn
from wsknn.preprocessing.static_parsers.json_parser.parse import parse_gzipped_fn, parse_jsonl_fn
from wsknn.preprocessing.structure.item import Items
from wsknn.preprocessing.structure.session import Sessions


def parse_files(dataset: str,
                session_id_key: str,
                product_key: str,
                time_key: str,
                action_key: str = None,
                time_to_numeric=False,
                time_to_datetime=False,
                datetime_format='',
                allowed_actions: Dict = None,
                purchase_action_name=None,
                progress_bar: bool = False) -> Tuple[Items, Sessions]:
    """
    Function parses data from csv, json and gzip json into item-sessions and session-items maps.

    Parameters
    ----------
    dataset : str
        The gzipped JSONL, JSON, CSV file with events.

    session_id_key : str
        The name of the session key.

    product_key : str
        The name of the product key.

    action_key : str
        The name of the event action type key.

    time_key : str
        The name of the event timestamp key.

    time_to_numeric : bool, default = False
        Transforms input timestamps to float values.

    time_to_datetime : bool, default = False
        Transforms input timestamps to datatime objects. Setting ``datetime_format`` parameter is required.

    datetime_format : str
        The format of datetime object.

    allowed_actions : Dict, optional
        Allowed actions and their weights.

    purchase_action_name: Any, optional
        The name of the final action (it is required to apply weight into the session vector).

    progress_bar : bool, default = False
        Show parsing progress.

    Returns
    -------
    items, sessions : Items, Sessions
        The mappings of item-session and session-items.
    """

    if dataset.endswith('.gz'):
        items, sessions = parse_gzipped_fn(dataset,
                                           allowed_actions,
                                           purchase_action_name,
                                           session_id_key,
                                           product_key,
                                           action_key,
                                           time_key,
                                           time_to_numeric,
                                           time_to_datetime,
                                           datetime_format,
                                           progress_bar)
    elif dataset.endswith('.json') or dataset.endswith('.jsonl'):
        items, sessions = parse_jsonl_fn(dataset,
                                         allowed_actions,
                                         purchase_action_name,
                                         session_id_key,
                                         product_key,
                                         action_key,
                                         time_key,
                                         time_to_numeric,
                                         time_to_datetime,
                                         datetime_format,
                                         progress_bar)
    elif dataset.endswith('.csv'):
        items, sessions = parse_csv_fn(dataset,
                                       allowed_actions,
                                       purchase_action_name,
                                       session_id_key,
                                       product_key,
                                       action_key,
                                       time_key,
                                       time_to_numeric,
                                       time_to_datetime,
                                       datetime_format,
                                       progress_bar)
    else:
        ftype = pathlib.Path(dataset).suffix
        raise TypeError(f'Unrecognized input file type. Parser works with "gz" (gzipped json), "json", and "csv"'
                        f'files, you have provided {ftype} type.')

    return items, sessions


def parse_flat_file(dataset: str,
                    sep: str,
                    session_index: int,
                    product_index: int,
                    time_index: int,
                    action_index: int = None,
                    use_header_row: bool = False,
                    time_to_numeric=False,
                    time_to_datetime=False,
                    datetime_format='',
                    allowed_actions: Dict = None,
                    purchase_action_name=None,
                    ignore_errors: bool = True):
    """
    Function parses data from flat file into item-sessions and session-items maps.

    Parameters
    ----------
    dataset : str
        Input file.

    sep : str
        Separator used to separate values.

    session_index : int
        The index of the session.

    product_index : int
        The index of the product.

    time_index : int
        The index of the event timestamp.

    action_index : int, optional
        The index of the event action.

    use_header_row : bool, default = False
        Use first row values as a header.

    time_to_numeric : bool, default = False
        Transforms input timestamps to float values.

    time_to_datetime : bool, default = False
        Transforms input timestamps to datatime objects. Setting ``datetime_format`` parameter is required.

    datetime_format : str
        The format of datetime object.

    allowed_actions : Dict, optional
        Allowed actions and their weights.

    purchase_action_name: Any, optional
        The name of the final action (it is required to apply weight into the session vector).

    ignore_errors : bool, default=True
        Ignore rows that raise exceptions.

    Returns
    -------
    items, sessions : Items, Sessions
        The mappings of item-session and session-items.
    """

    parsed_items, parsed_sessions = parse_flat_file_fn(
        dataset=dataset,
        sep=sep,
        session_index=session_index,
        product_index=product_index,
        time_index=time_index,
        action_index=action_index,
        use_header_row=use_header_row,
        time_to_numeric=time_to_numeric,
        time_to_datetime=time_to_datetime,
        datetime_format=datetime_format,
        allowed_actions=allowed_actions,
        purchase_action_name=purchase_action_name,
        ignore_errors=ignore_errors
    )

    return parsed_items, parsed_sessions



import gzip
import json

import numpy as np
import tqdm


def simulate_input(number_of_sessions: int,
                   number_of_items: int,
                   min_session_length: int,
                   max_session_length: int,
                   output_sessions_json_path: str,
                   output_items_json_path: str,
                   item_occurence_distribution="uniform",
                   compress=True):
    """
    Function generates fake sessions and items for testing purposes.

    Parameters
    ----------
    number_of_sessions : int
        The number of sessions.

    number_of_items : int
        The number of items.

    min_session_length : int
        The minimum session length.

    max_session_length : int
        The maximum session length.

    output_sessions_json_path : str
        The path to the output sessions file.

    output_items_json_path : str
        The path to the output items file.

    item_occurence_distribution : str
        How items are sampled, for now the only possible distribution is "uniform".

    compress : bool, default=True
        Should jsonl be compressed to gzip?
    """

    items_map = {}
    sessions_map = {}

    for session_idx in tqdm.tqdm(range(0, number_of_sessions)):
        session_length = np.random.randint(min_session_length, max_session_length+1)
        items = np.random.randint(0, high=number_of_items, size=session_length)
        tstart = np.random.randint(1, 10)
        tend = tstart * max_session_length
        times = np.linspace(tstart, tend, session_length)

        sessions_map[session_idx] = [
            items.tolist(), times.tolist()
        ]

        for item in items:
            item = int(item)
            if item in items_map:
                items_map[item][0].append(session_idx)
                items_map[item][1].append(times[0])
            else:
                items_map[item] = [
                    [session_idx], [times[0]]
                ]

    if compress:
        output_sessions_json_path = output_sessions_json_path + '.gz'
        with gzip.open(output_sessions_json_path, 'wt', encoding='UTF-8') as zipfile:
            json.dump(sessions_map, zipfile)

        output_items_json_path = output_items_json_path + '.gz'
        with gzip.open(output_items_json_path, 'wt', encoding='UTF-8') as zipfile:
            json.dump(items_map, zipfile)

    del sessions_map
    del items_map

    return output_items_json_path, output_sessions_json_path


# if __name__ == '__main__':
#
#     out_dir = ''
#     item_sizes = np.arange(1000, 112000, 10000)
#     item_sizes = [int(x) for x in item_sizes]
#     session_sizes = [100_000]
#
#     for sess_size in tqdm.tqdm(session_sizes):
#         for item_size in tqdm.tqdm(item_sizes):
#             sess_size = int(sess_size)
#             item_size = int(item_size)
#             out_sessions = out_dir + f'smap_s_{sess_size}_i_{item_size}.json'
#             out_items = out_dir + f'imap_s_{sess_size}_i_{item_size}.json'
#             simulate_input(number_of_sessions=sess_size,
#                            number_of_items=item_size,
#                            min_session_length=1,
#                            max_session_length=30,
#                            output_sessions_json_path=out_sessions,
#                            output_items_json_path=out_items)



import unittest
from wsknn.preprocessing.utils.transform import parse_seconds_to_dt, parse_dt_to_seconds


INPUT_DATE_LONG = '2021-12-05T00:00:00.000Z'
INPUT_DATE_SHORT = '2021-12-05T00:00:00'
INPUT_DATE_TIMESTAMP = 1638658800
WRONG_INPUT = '2021-01-01'


class TestParseTime(unittest.TestCase):

    def test_dt_to_seconds(self):
        output_long = parse_dt_to_seconds(INPUT_DATE_LONG)
        output_short = parse_dt_to_seconds(INPUT_DATE_SHORT)
        self.assertEqual(output_short, output_long)
        self.assertRaises(TypeError, parse_dt_to_seconds, WRONG_INPUT)

    def test_seconds_to_dt(self):

        output = parse_seconds_to_dt(INPUT_DATE_TIMESTAMP)
        self.assertIsInstance(output, str)


if __name__ == '__main__':
    unittest.main()



from typing import Dict, List
import pandas as pd
from wsknn.preprocessing.structure.session_to_item_map import map_sessions_to_items


def _clip_multiple_transactions(session, items, times, actions, locs, min_session_length):
    loclen = len(items) - 1

    parsed = []
    pidx = None

    for idx, lc in enumerate(locs):
        if lc == 0:
            continue
        else:
            if idx == 0:
                previous = 0
            else:
                previous = locs[idx-1] + 1

            if lc + 1 < loclen:
                lc = lc + 1

            nitem = items[previous:lc]
            ntime = times[previous:lc]
            nact = actions[previous:lc]
            if len(nitem) >= min_session_length:
                parsed.append(
                    [session, nitem, ntime, nact]
                )
        pidx = idx

    if pidx < loclen:
        nitem = items[pidx:]
        ntime = times[pidx:]
        nact = actions[pidx:]
        if len(nitem) >= min_session_length:
            parsed.append(
                [session, nitem, ntime, nact]
            )

    return parsed


def _parse_full_ds(sess_idx, items, times, actions, is_transaction, min_session_length):
    if is_transaction is None:
        # Skip dividing sessions and put everything in one sequence
        parsed = {
            x: [items[idx], times[idx], actions[idx]]
            for idx, x in enumerate(sess_idx) if len(items[idx]) >= min_session_length}
    else:
        parsed = {}
        for idx, x in enumerate(is_transaction):
            if 1 in x:
                if len(items[idx]) > min_session_length:
                    parsed[sess_idx[idx]] = [items[idx], times[idx], actions[idx]]
            # TODO: complex parsing schema, where single session may occur multiple times! FUTURE
            #     # We have a transaction here, time to divide sequence
            #     t_idx = list(locate(x))
            #     # Divide sequences based on the transaction indices
            #     parsed.extend(
            #         _clip_multiple_transactions(
            #             session=sess_idx[idx], items=items[idx], times=times[idx], actions=actions[idx], locs=t_idx,
            #             min_session_length=min_session_length
            #         )
            #     )
            # else:
            #     if len(items[idx]) >= min_session_length:
            #         parsed.append(
            #             {sess_idx[idx]: [items[idx], times[idx], actions[idx]]}
            #         )
    return parsed


def _prepare_values_session_map(df: pd.DataFrame,
                                session_id_key: str,
                                product_key: str,
                                time_key: str,
                                action_key: str = None,
                                purchase_action_name: str = None):
    df = df.sort_values([session_id_key, time_key])
    gdf = df.groupby(session_id_key)
    timestamps = gdf[time_key].apply(list).values
    products = gdf[product_key].apply(list)
    actions = None
    transactions = None
    if action_key is not None:
        actions = gdf[action_key].apply(list).values
    if purchase_action_name is not None:
        transactions = gdf['is_transaction'].apply(list).values

    return products.index, products.values, timestamps, actions, transactions


def _build_maps_from_df(df: pd.DataFrame,
                        session_id_key: str,
                        product_key: str,
                        time_key: str,
                        action_key: str = None,
                        purchase_action_name: str = None,
                        min_session_length: int = 3) -> Dict:
    # Prepare data for session map
    sess_idx, items, times, actions, is_transaction = _prepare_values_session_map(
        df,
        session_id_key,
        product_key,
        time_key,
        action_key,
        purchase_action_name
    )

    if action_key is None:
        level_1_parsed = {
            x: [items[idx], times[idx]]
            for idx, x in enumerate(sess_idx) if len(items[idx]) >= min_session_length
        }
    else:
        level_1_parsed = _parse_full_ds(sess_idx, items, times, actions, is_transaction, min_session_length)

    return level_1_parsed


def parse_pandas(df: pd.DataFrame,
                 session_id_key: str,
                 product_key: str,
                 time_key: str,
                 action_key: str = None,
                 allowed_actions: List = None,
                 purchase_action_name: str = None,
                 min_session_length: int = 3,
                 get_items_map: bool = True) -> Dict:
    """
    Function parses given dataset into Sessions and Items objects.

    Parameters
    ----------
    df : pandas DataFrame
        Dataframe with events and sessions.

    session_id_key : str
        The name of the session key.

    product_key : str
        The name of the product key.

    time_key : str
        The name of the event timestamp key.

    action_key : str, default = None
        The name of the event action type key.

    allowed_actions : List, optional
        Allowed actions.

    purchase_action_name: Any, optional
        The name of the final action (it is required to apply weight into the session vector).

    min_session_length : int, default = 3
        Minimum length of a single session.

    get_items_map : bool, default = True
        Should item-sessions map be created?

    Returns
    -------
    : Dict
        {"session-map": Dict, "item-map": Optional[Dict]}
    """

    # Clean dataframe
    if action_key is not None:
        # Check allowed actions
        if allowed_actions is not None:
            # Select rows with allowed actions
            df = df[df[action_key].isin(allowed_actions)]
        # Check purchase action
        if purchase_action_name is not None:
            df['is_transaction'] = (df[action_key] == purchase_action_name).astype(int)

    # Prepare maps
    sess_map = _build_maps_from_df(
        df,
        session_id_key,
        product_key,
        time_key,
        action_key,
        purchase_action_name,
        min_session_length
    )

    output = {'session-map': sess_map}

    if get_items_map:
        items_map = map_sessions_to_items(sessions_map=sess_map)
        output['item-map'] = items_map

    return output



import unittest
from wsknn.preprocessing.utils.calc import get_larger_value


class TestGetLargerValue(unittest.TestCase):

    def test_int_inputs(self):
        INPUT1 = 10
        INPUT2 = 0
        INPUT3 = -10

        EXPECTED_OUTPUT_1 = 10
        EXPECTED_OUTPUT_2 = 0
        EXPECTED_OUTPUT_3 = 0

        t1 = get_larger_value(INPUT1, INPUT2)
        t2 = get_larger_value(INPUT2, INPUT2)
        t3 = get_larger_value(INPUT2, INPUT3)

        # Simple test
        self.assertEqual(t1, EXPECTED_OUTPUT_1)
        self.assertEqual(t2, EXPECTED_OUTPUT_2)
        self.assertEqual(t3, EXPECTED_OUTPUT_3)

    def test_float_inputs(self):
        INPUT1 = 10.0
        INPUT2 = 0.0
        INPUT3 = 10.001

        EXPECTED_OUTPUT_1 = 10.001
        EXPECTED_OUTPUT_2 = 0.0

        t1 = get_larger_value(INPUT1, INPUT3)
        t2 = get_larger_value(INPUT2, INPUT2)

        # Simple test
        self.assertEqual(t1, EXPECTED_OUTPUT_1)
        self.assertEqual(t2, EXPECTED_OUTPUT_2)


if __name__ == '__main__':
    unittest.main()



from typing import Dict

import numpy as np
import pandas as pd

from wsknn.preprocessing.static_parsers.pandas_parser import parse_pandas


df = pd.read_csv('tdata/ml-25m-test/ratings.csv')

_ = parse_pandas(
    df=df,
    session_id_key='userId',
    product_key='movieId',
    time_key='timestamp'
)

possible_actions = ['ORDER', 'VIEW']
actions_probas = [0.05, 0.95]

actions = np.random.choice(possible_actions, size=len(df), p=actions_probas)
df['action'] = actions

def test_parsing():
    parsed = parse_pandas(
        df=df,
        session_id_key='userId',
        product_key='movieId',
        time_key='timestamp',
        action_key='action',
        get_items_map=False
    )

    assert isinstance(parsed, Dict)
    assert parsed.get('session-map')
    assert not parsed.get('item-map')

    parsed = parse_pandas(
        df=df,
        session_id_key='userId',
        product_key='movieId',
        time_key='timestamp',
        action_key='action',
        allowed_actions=['ORDER']
    )

    assert isinstance(parsed, Dict)
    assert parsed.get('session-map')
    assert parsed.get('item-map')

    parsed = parse_pandas(
        df=df,
        session_id_key='userId',
        product_key='movieId',
        time_key='timestamp',
        action_key='action',
        purchase_action_name='ORDER',
        min_session_length=3
    )

    assert isinstance(parsed, Dict)
    assert parsed.get('session-map')
    assert parsed.get('item-map')



from typing import Dict
import csv

from wsknn.preprocessing.static_parsers.parse import parse_fn


def parse_csv_fn(dataset: str,
                 allowed_actions: Dict,
                 purchase_action_name: str,
                 session_id_key: str,
                 product_key: str,
                 action_key: str,
                 time_key: str,
                 time_to_numeric: bool,
                 time_to_datetime: bool,
                 datetime_format: str,
                 progress_bar: bool):
    """
    Function parses given CSV file into Sessions and Items and Users objects.

    Parameters
    ----------
    dataset : str
        The JSON file with events.

    allowed_actions : Dict, optional
        Allowed actions and their weights.

    purchase_action_name: Any, optional
        The name of the final action (it is required to apply weight into the session vector).

    session_id_key : str
        The name of the session key (column header).

    product_key : str
        The name of the product key (column header).

    action_key : str
        The name of the event action type key (column header).

    time_key : str
        The name of the event timestamp key (column header).

    time_to_numeric : bool, default = True
        Transforms input timestamps to float values.

    time_to_datetime : bool, default = False
        Transforms input timestamps to datatime objects. Setting ``datetime_format`` parameter is required.

    datetime_format : str
        The format of datetime object.

    progress_bar : bool
        Show parsing progress.

    Returns
    -------
    ItemsMap, SessionsMap : Items, Sessions
    """
    with open(dataset, 'r', encoding='utf-8') as csvdata:
        try:
            csvstream = csv.DictReader(csvdata)
        except Exception as _:
            raise IOError(f'Cannot read {dataset} file!')

        parsed_items, parsed_sessions = parse_fn(
            csvstream,
            allowed_actions,
            purchase_action_name,
            session_id_key,
            product_key,
            action_key,
            time_key,
            time_to_numeric,
            time_to_datetime,
            datetime_format,
            progress_bar
        )

    return parsed_items, parsed_sessions



import os
import yaml
from datetime import datetime, timedelta
from pathlib import Path


PREPROCESSING_FILENAME = 'preprocessing_settings.yml'
IO_KEY = 'io'
PROPERTIES_KEY = 'properties'
DATE_FORMAT = '%Y-%m-%d'


def get_files_list(directory: str, ftype: str) -> list:
    """
    Function returns list of files with a given type.

    Parameters
    ----------
    directory : str
        The path to directory with files.

    ftype : str
        File type.

    Returns
    -------
    files : List
        The list with files of specified type.
    """
    files = [os.path.join(directory, x) for x in os.listdir(directory) if x.endswith(ftype)]
    return files


def get_files_list_from_daterange(directory: str, ftype: str, date_start=None, date_end=None) -> list:
    """
    Function returns list of selected file types from a given directory within a specified dates. Filename MUST contain
    date in the format YYYY-MM-DD.

    Parameters
    ----------
    directory : str
        The path to directory with files.

    ftype : str
        File type.

    date_start : str, optional
        Date in ISO format YYYY-MM-DD. If you provide ``start_date`` then files in the past from this date won't
        be included.

    date_end : str, optional
        Date in ISO format YYYY-MM-DD. If you provide ``end_date`` then files in the future from this date won't
        be included.

    Returns
    -------
    files : List
        The list with files of specified type within date ranges.
    """
    files = get_files_list(directory, ftype)
    if date_start is not None or date_end is not None:
        files = _filter_files_by_date(files, date_start, date_end)

    # Sort files
    files = sorted(files)
    return files


def _filter_files_by_date(files, date_start, date_end):
    if date_start is None:
        possible_dates = _get_possible_dates_upper_bound_only(date_end, files)
    elif date_end is None:
        possible_dates = _get_possible_dates_lower_bound_only(date_start, files)
    else:
        possible_dates = _get_possible_dates(date_start, date_end)

    files_new = []
    for pdate in possible_dates:
        for fname in files:
            if pdate in fname:
                files_new.append(fname)

    return files_new


def _get_possible_dates_upper_bound_only(upper_date, fileslist):
    sorted_files = sorted(fileslist)
    flist = []
    check = 0
    for fname in sorted_files:
        if check == 0:
            flist.append(fname)
            if upper_date in fname:
                check = 1
        else:
            if upper_date in fname:
                flist.append(fname)
            else:
                break

    return flist


def _get_possible_dates_lower_bound_only(lower_date, fileslist):
    sorted_files = sorted(fileslist)
    flist = []
    check = 0
    for fname in sorted_files:
        if check == 0:
            if lower_date in fname:
                flist.append(fname)
                check = 1
        else:
            flist.append(fname)

    return flist


def _get_possible_dates(lower_date, upper_date):
    date_low = datetime.strptime(lower_date, DATE_FORMAT)

    dates = [lower_date]
    d0 = date_low
    while True:
        d0 = d0 + timedelta(days=1)
        dstr = d0.strftime(DATE_FORMAT)
        dates.append(dstr)
        if dstr == upper_date:
            break

    return dates


def search_upwards_for_file(filename: str):
    """
    Search in the current directory and all directories above it
    for a file of a particular name.

    Parameters
    ----------
    filename : str
        The filename to look for.

    Returns
    -------
    : pathlib.Path
        The location of the first file found or ``None``, if none was found
    """
    d = Path.cwd()
    root = Path(d.root)

    while d != root:
        attempt = d / filename
        if attempt.exists():
            return attempt
        d = d.parent

    return None


def parse_settings(fpath=None):
    """
    Function parses ``preprocessing_settings`` file.

    Parameters
    ----------
    fpath : str, optional
        If given settings are retrieved from file, otherwise function tries to find ``preprocessing_settings`` in a
        working directory.

    Returns
    -------
    : Dict
        Parsed settings.

    Raises
    ------
    RunetimeError
        File not found or it has missing keys.
    """

    if fpath is None:
        settings_file = PREPROCESSING_FILENAME
        path_to_file = search_upwards_for_file(settings_file)
    else:
        path_to_file = fpath

    with open(path_to_file, 'r') as settings_stream:
        parsed_settings = yaml.safe_load(settings_stream)

    settings_keys = list(parsed_settings.keys())
    if IO_KEY in settings_keys:
        if PROPERTIES_KEY in settings_keys:
            return parsed_settings

    msg = f'Does the {PREPROCESSING_FILENAME} file exist in a top level of the module? ' \
          f'Does it have "{IO_KEY}" and "{PROPERTIES_KEY}" keys?'
    raise RuntimeError(msg)



from typing import Dict
from more_itertools import sort_together


def map_sessions_to_items(sessions_map: Dict, sort_items_map: bool = True) -> Dict:
    """
    Function transforms sessions map into items map.

    Parameters
    ----------
    sessions_map : Dict
        >>> sessions = {
        ...    session_id: (
        ...        [ sequence_of_items ],
        ...        [ sequence_of_timestamps ],
        ...        [ [OPTIONAL] sequence_of_event_types ],
        ...        [ [OPTIONAL] sequence_of_event_weights]
        ...    )
        ...}

    sort_items_map : bool, default = True
        Sorts item-sessions map by the first timestamp of a session (then sessions are sorted in ascending order, from
        the oldest to the newest).

    Returns
    -------
    items_map : Dict
        Mapped item-sessions dictionary.
        >>> items = {
        ...    item_id: (
        ...        [ sequence_of_sessions ],
        ...        [ sequence_of_the_first_session_timestamps ]
        ...    )
        ...}
    """

    items_map = {}

    for session_k, values in sessions_map.items():

        items = values[0]
        timestamps = values[1]

        for idx, item in enumerate(items):
            ts = timestamps[idx]
            if item not in items_map:
                items_map[item] = [
                    [session_k],
                    [ts]
                ]
            else:
                if session_k not in items_map[item][0]:
                    items_map[item][0].append(session_k)
                    items_map[item][1].append(ts)
                else:
                    session_index = items_map[item][0].index(session_k)
                    the_last_timestamp = items_map[item][1][session_index]
                    if ts < the_last_timestamp:
                        items_map[item][1][session_index] = ts

    if sort_items_map:
        for ikey, ivalues in items_map.items():
            # sort values
            new_ivalues = list(sort_together(ivalues, key_list=(1,)))
            # update mapping
            items_map[ikey] = new_ivalues

    return items_map



from wsknn import parse_files
from wsknn.preprocessing.structure.session_to_item_map import map_sessions_to_items

JSONFILE = 'tdata/events1.json'
SESSION_KEY = 'sid'
PRODUCT_KEY = 'pid'
ACTION_KEY = 'act'
TIME_KEY = 'ts'
POSSIBLE_ACTIONS = {'products_view': 0.1,
                    'purchase': 1,
                    'add_to_wishlist': 0.3,
                    'add_to_cart': 0.5}
PURCHASE_ACTION_NAME = 'purchase'


def test_get_items_map_from_sessions_map():
    parsed_items, parsed_sessions = parse_files(
        dataset=JSONFILE,
        session_id_key=SESSION_KEY,
        product_key=PRODUCT_KEY,
        action_key=ACTION_KEY,
        time_key=TIME_KEY,
        time_to_numeric=True,
        allowed_actions=POSSIBLE_ACTIONS,
        purchase_action_name=PURCHASE_ACTION_NAME
    )
    sessions_map = parsed_sessions.session_items_actions_map
    items_map = parsed_items.item_sessions_map

    items_from_sessions_map = map_sessions_to_items(sessions_map, sort_items_map=False)

    for original_key, original_mapping in items_map.items():
        assert original_key in items_from_sessions_map
        sessions, timestamps = original_mapping[0], original_mapping[1]
        for idx, sess in enumerate(sessions):
            assert sess in items_from_sessions_map[original_key][0]
            assert timestamps[idx] in items_from_sessions_map[original_key][1]



import datetime
from typing import Union, List, Dict
from more_itertools import sort_together


def merge_dicts(dict1: Dict, dict2: Dict) -> Dict:
    """
    Function merges two dicts.

    Parameters
    ----------
    dict1 : Dict

    dict2 : Dict

    Returns
    -------
    : Dict
    """
    merged = dict1.copy()

    for key_d2 in dict2.keys():
        d2vals = dict2[key_d2]
        if key_d2 in dict1:
            d1vals = dict1[key_d2]
            vals = merge_vals(d1vals, d2vals)
            merged[key_d2] = vals
        else:
            merged[key_d2] = d2vals

    return merged


def merge_vals(vals1: List, vals2: List) -> List:
    """
    Function merges lists of values and leave only unique elements.

    Parameters
    ----------
    vals1 : List
        The list with values to merge.

    vals2 : List
        The list with values to merge.

    Returns
    -------
    : List
        Merged and lists.
    """

    vals1[0].extend(vals2[0])
    vals1[1].extend(vals2[1])

    if len(vals2) >= 3:
        vals1[2].extend(vals2[2])

    if len(vals2) == 4:
        vals1[3].extend(vals2[3])

    # Sort by time
    vals1 = sort_together(vals1, key_list=(1,))
    vals_output = []
    for val_row in vals1:
        if isinstance(val_row, tuple):
            vals_output.append(list(val_row))
        else:
            vals_output.append(val_row)

    return vals_output


def parse_dt_to_seconds(event_time: Union[int, str]) -> int:
    """
    Function converts given date to seconds from 1970-01-01. Function rounds returned value to a full second.

    Parameters
    ----------
    event_time : Union[int, str]
        Even time as timestamp or string. Datetime in formats '%Y-%m-%dT%H:%M:%S.%fZ' or '%Y-%m-%dT%H:%M:%S'

    Returns
    -------
    timestamp_ticks : int
        How many seconds to ``event_time`` from 1970-01-01, UTC tz.
    """
    try:
        int(event_time)
        return event_time
    except ValueError:
        if isinstance(event_time, str):
            if len(event_time) >= 24:
                str_scheme = '%Y-%m-%dT%H:%M:%S.%fZ'
            elif len(event_time) == 19:
                str_scheme = '%Y-%m-%dT%H:%M:%S'
            else:
                raise TypeError('Given Datetime Format not allowed! Function uses "%Y-%m-%dT%H:%M:%S.%fZ" and '
                                '"%Y-%m-%dT%H:%M:%S" formats')
            base_time = datetime.datetime.strptime(event_time, str_scheme)
            # Transform to utc

            timestamp_ticks = int(base_time.timestamp())
            return timestamp_ticks
        else:
            raise TypeError(f'Unrecognized datetime format: {type(event_time)}')


def parse_seconds_to_dt(timestamp2transform: int) -> str:
    """
    Function converts given timestamp in seconds to string datetime in format '%Y-%m-%dT%H:%M:%S.%fZ'.

    Parameters
    ----------
    timestamp2transform : int
        Timestamp in seconds to transform.

    Returns
    -------
    s_datetime : str
        Parsed datetime in format '%Y-%m-%dT%H:%M:%S.%fZ'
    """
    s_datetime = datetime.datetime.fromtimestamp(
        timestamp2transform
    ).strftime('%Y-%m-%dT%H:%M:%S.%fZ')

    return s_datetime


def parse_event_actions(action: str, all_actions: List) -> Dict:
    """
    Function maps event actions to a vector of actions.

    :param action: (str) name of the user action
    :param all_actions: (list) available actions
    :return: (Dict) {action: 1, other_action: 0} where action is equal to given action parameter and other_action
             is other action type from the all_actions_list.
    """

    actions_dict = {}

    for act in all_actions:
        if act == action:
            actions_dict[act] = 1
        else:
            actions_dict[act] = 0

    return actions_dict



import unittest

from wsknn.preprocessing.parse_static import parse_files
from wsknn.preprocessing.structure.item import Items
from wsknn.preprocessing.structure.session import Sessions

POSSIBLE_ACTIONS = {'products_view': 0.1,
                    'purchase': 1,
                    'add_to_wishlist': 0.3,
                    'add_to_cart': 0.5}

PURCHASE_ACTION_NAME = 'purchase'
SESSIONS = ['tdata/events1.json',
            'tdata/events2.json',
            'tdata/events3.json']

SESSION_KEY = 'sid'
PRODUCT_KEY = 'pid'
ACTION_KEY = 'act'
TIME_KEY = 'ts'


class TestParseFn(unittest.TestCase):

    def test_instances(self):

        for fs in SESSIONS:
            out_item, out_session = parse_files(fs,
                                                session_id_key=SESSION_KEY,
                                                product_key=PRODUCT_KEY,
                                                action_key=ACTION_KEY,
                                                time_key=TIME_KEY,
                                                allowed_actions=POSSIBLE_ACTIONS,
                                                purchase_action_name=PURCHASE_ACTION_NAME,
                                                time_to_numeric=True)

            self.assertIsInstance(out_item, Items)
            self.assertIsInstance(out_session, Sessions)

    def test_output_values(self):

        base_items, base_sessions = parse_files(SESSIONS[0],
                                                session_id_key=SESSION_KEY,
                                                product_key=PRODUCT_KEY,
                                                action_key=ACTION_KEY,
                                                time_key=TIME_KEY,
                                                allowed_actions=POSSIBLE_ACTIONS,
                                                purchase_action_name=PURCHASE_ACTION_NAME,
                                                time_to_numeric=True)

        base_items_keys = list(base_items.item_sessions_map.keys()).copy()
        base_sessions_keys = list(base_sessions.session_items_actions_map.keys()).copy()

        other_items_keys = [base_items_keys]
        other_sessions_keys = [base_sessions_keys]

        for fs in SESSIONS[1:]:
            out_item, out_session = parse_files(fs,
                                                allowed_actions=POSSIBLE_ACTIONS,
                                                purchase_action_name=PURCHASE_ACTION_NAME,
                                                session_id_key=SESSION_KEY,
                                                product_key=PRODUCT_KEY,
                                                action_key=ACTION_KEY,
                                                time_key=TIME_KEY,
                                                time_to_numeric=True,
                                                time_to_datetime=False,
                                                datetime_format='')
            base_items = base_items + out_item
            base_sessions = base_sessions + out_session

            # Append keys
            other_items_keys.append(
                list(out_item.item_sessions_map.keys())
            )
            other_sessions_keys.append(
                list(out_session.session_items_actions_map.keys())
            )

        # Check items keys
        ikeys = set(base_items.item_sessions_map.keys())
        for ik in other_items_keys:
            ikset = set(ik)
            isik = ikset.issubset(ikeys)
            msg = 'Missing keys in items-sessions map dictionary!'
            self.assertTrue(isik, msg)

        # Check sessions keys
        skeys = set(base_sessions.session_items_actions_map.keys())
        for sk in other_sessions_keys:
            skset = set(sk)
            issk = skset.issubset(skeys)
            msg = 'Missing keys in sessions-items map dictionary!'
            self.assertTrue(issk, msg)

    def test_session_weighting(self):
        sample_session = Sessions(event_session_key='sid',
                                  event_product_key='pid',
                                  event_time_key='dt',
                                  event_action_key='action',
                                  event_action_weights={
                                      'products_view': 0.1,
                                      'purchase': 1
                                  })
        event1 = {
            'sid': '0',
            'pid': 'a',
            'dt': 1,
            'action': 'products_view'
        }
        event2 = {
            'sid': '0',
            'pid': 'b',
            'dt': 2,
            'action': 'products_view'
        }
        event3 = {
            'sid': '0',
            'pid': 'c',
            'dt': 3,
            'action': 'products_view'
        }
        event4 = {
            'sid': '0',
            'dt': 4,
            'action': 'purchase'
        }

        event_4_products = ['a', 'c']

        sample_session.append(event1)
        sample_session.append(event2)
        sample_session.append(event3)
        sample_session.update_weights_of_purchase_session(event4['sid'], 1, bought_products=event_4_products)

        expected_output = {
            '0': (
                ['a', 'b', 'c'],
                [1, 2, 3],
                ['products_view', 'products_view', 'products_view'],
                [1.1, 0.1, 1.1]
            )
        }
        self.assertEqual(sample_session.session_items_actions_map, expected_output)

    def test_none_action(self):

        for fs in SESSIONS:
            out_item, out_session = parse_files(fs,
                                                session_id_key=SESSION_KEY,
                                                product_key=PRODUCT_KEY,
                                                time_key=TIME_KEY,
                                                action_key=None,
                                                allowed_actions=None,
                                                purchase_action_name=None,
                                                time_to_numeric=True)

            self.assertIsInstance(out_item, Items)
            self.assertIsInstance(out_session, Sessions)


if __name__ == '__main__':
    unittest.main()



from wsknn import parse_files, Items, Sessions


JSONFILE = 'tdata/events1.json'
GZJSONFILE = 'tdata/events1.json.gz'
CSVFILE = 'tdata/events1.csv'
SESSION_KEY = 'sid'
PRODUCT_KEY = 'pid'
ACTION_KEY = 'act'
TIME_KEY = 'ts'
POSSIBLE_ACTIONS = {'products_view': 0.1,
                    'purchase': 1,
                    'add_to_wishlist': 0.3,
                    'add_to_cart': 0.5}
PURCHASE_ACTION_NAME = 'purchase'


def test_parse_json():
    parsed_items, parsed_sessions = parse_files(
        dataset=JSONFILE,
        session_id_key=SESSION_KEY,
        product_key=PRODUCT_KEY,
        action_key=ACTION_KEY,
        time_key=TIME_KEY,
        time_to_numeric=True,
        allowed_actions=POSSIBLE_ACTIONS,
        purchase_action_name=PURCHASE_ACTION_NAME
    )

    assert isinstance(parsed_items, Items)
    assert isinstance(parsed_sessions, Sessions)


def test_parse_csv():
    parsed_items, parsed_sessions = parse_files(
        dataset=CSVFILE,
        session_id_key=SESSION_KEY,
        product_key=PRODUCT_KEY,
        action_key=ACTION_KEY,
        time_key=TIME_KEY,
        time_to_numeric=True,
        allowed_actions=POSSIBLE_ACTIONS,
        purchase_action_name=PURCHASE_ACTION_NAME
    )

    assert isinstance(parsed_items, Items)
    assert isinstance(parsed_sessions, Sessions)


def test_parse_gzip_json():
    parsed_items, parsed_sessions = parse_files(
        dataset=GZJSONFILE,
        session_id_key=SESSION_KEY,
        product_key=PRODUCT_KEY,
        action_key=ACTION_KEY,
        time_key=TIME_KEY,
        time_to_numeric=True,
        allowed_actions=POSSIBLE_ACTIONS,
        purchase_action_name=PURCHASE_ACTION_NAME
    )

    assert isinstance(parsed_items, Items)
    assert isinstance(parsed_sessions, Sessions)



import gzip
import json
from typing import Dict

from wsknn.preprocessing.static_parsers.parse import parse_fn


def parse_gzipped_fn(dataset: str,
                     allowed_actions: Dict,
                     purchase_action_name: str,
                     session_id_key: str,
                     product_key: str,
                     action_key: str,
                     time_key: str,
                     time_to_numeric: bool,
                     time_to_datetime: bool,
                     datetime_format: str,
                     progress_bar: bool):
    """
    Function parses given gzipped JSONL file into Sessions and Items objects.

    Parameters
    ----------
    dataset : str
        The gzipped JSONL file with events.

    allowed_actions : Dict, optional
        Allowed actions and their weights.

    purchase_action_name: Any, optional
        The name of the final action (it is required to apply weight into the session vector).

    session_id_key : str
        The name of the session key.

    product_key : str
        The name of the product key.

    action_key : str
        The name of the event action type key.

    time_key : str
        The name of the event timestamp key.

    time_to_numeric : bool, default = True
        Transforms input timestamps to float values.

    time_to_datetime : bool, default = False
        Transforms input timestamps to datatime objects. Setting ``datetime_format`` parameter is required.

    datetime_format : str
        The format of datetime object.

    progress_bar : bool
        Show parsing progress.

    Returns
    -------
    parsed_items, parsed_sessions : Items, Sessions
        The mappings of item-session and session-items.
    """
    with gzip.open(dataset, 'rt', encoding='UTF-8') as unzipped_f:

        try:
            jstream = json.load(unzipped_f)
        except json.decoder.JSONDecodeError:
            jstream = [json.loads(x) for x in unzipped_f]

        parsed_items, parsed_sessions = parse_fn(
            jstream,
            allowed_actions,
            purchase_action_name,
            session_id_key,
            product_key,
            action_key,
            time_key,
            time_to_numeric,
            time_to_datetime,
            datetime_format,
            progress_bar
        )

    return parsed_items, parsed_sessions


def parse_jsonl_fn(dataset: str,
                   allowed_actions: Dict,
                   purchase_action_name: str,
                   session_id_key: str,
                   product_key: str,
                   action_key: str,
                   time_key: str,
                   time_to_numeric: bool,
                   time_to_datetime: bool,
                   datetime_format: str,
                   progress_bar: bool):
    """
    Function parses given JSONL file into Sessions and Items and Users objects.

    Parameters
    ----------
    dataset : str
        The JSON file with events.

    allowed_actions : Dict, optional
        Allowed actions and their weights.

    purchase_action_name: Any, optional
        The name of the final action (it is required to apply weight into the session vector).

    session_id_key : str
        The name of the session key.

    product_key : str
        The name of the product key.

    action_key : str
        The name of the event action type key.

    time_key : str
        The name of the event timestamp key.

    time_to_numeric : bool, default = True
        Transforms input timestamps to float values.

    time_to_datetime : bool, default = False
        Transforms input timestamps to datatime objects. Setting ``datetime_format`` parameter is required.

    datetime_format : str
        The format of datetime object.

    progress_bar : bool
        Show parsing progress.

    Returns
    -------
    ItemsMap, SessionsMap : Items, Sessions
    """
    with open(dataset, 'r', encoding='utf-8') as jsondata:
        try:
            jstream = json.load(jsondata)
        except json.decoder.JSONDecodeError:
            jstream = [json.loads(x) for x in jsondata]

        parsed_items, parsed_sessions = parse_fn(
            jstream,
            allowed_actions,
            purchase_action_name,
            session_id_key,
            product_key,
            action_key,
            time_key,
            time_to_numeric,
            time_to_datetime,
            datetime_format,
            progress_bar
        )

    return parsed_items, parsed_sessions



import gzip
import json


def to_json(datadict: dict, filename: str, compress=True) -> None:
    """
    Function saves mapped object into jsonl file.

    Parameters
    ----------
    datadict : Dict
        Data to be stored.

    filename : str
        The path to the stored object. If suffix ``.json`` or ``.jsonl`` is not given then methods appends
        ``.json`` suffix into the file.

    compress : bool, default=True
        Should file be compressed into a gzip archive?
    """

    sjson = '.json'
    sjsonl = '.jsonl'
    sgz = '.gz'

    if (not filename.endswith(sjson)) or (not filename.endswith(sjsonl)):
        filename = filename + sjson

    if compress:
        filename = filename + sgz
        with gzip.open(filename, 'w') as compressed:
            for kdx, vals in datadict.items():
                d = {kdx: vals}
                jout = json.dumps(d) + '\n'
                jout = jout.encode('utf-8')
                compressed.write(jout)
    else:
        with open(filename, 'w') as out:
            for kdx, vals in datadict.items():
                d = {kdx: vals}
                jout = json.dumps(d) + '\n'
                out.write(jout)



import unittest
from wsknn.preprocessing.structure.session import Sessions


SESSION_KEY = 'sid'
ITEM_KEY = 'pid'
T_KEY = 'dt'
ACTION_KEY = 'action'
ACTION_WEIGHTS = {'products_view': 0.1,
                  'purchase': 1,
                  'add_to_wishlist': 0.3,
                  'add_to_cart': 0.5}

EVENT1 = {
    'sid': '0',
    'pid': 'a',
    'dt': 1,
    'action': 'products_view'
}

EVENT2 = {
    'sid': '0',
    'pid': 'b',
    'dt': 2,
    'action': 'products_view'
}

EVENT3 = {
    'sid': '0',
    'pid': 'c',
    'dt': 3,
    'action': 'products_view'
}

EVENT4 = {
    'sid': '0',
    'dt': 4,
    'action': 'purchase'
}

EVENT_4_PRODUCTS = ['a', 'c']

# TODO: update Sessions class test
class TestSessionsClass(unittest.TestCase):

    def test_init(self):
        sessions = Sessions(event_session_key=SESSION_KEY,
                            event_product_key=ITEM_KEY,
                            event_time_key=T_KEY,
                            event_action_key=ACTION_KEY,
                            event_action_weights=ACTION_WEIGHTS)

        self.assertIsInstance(sessions, Sessions)

        EXPECTED_MAP = dict()
        EXPECTED_TIME_START = 1_000_000_000_000_000
        EXPECTED_TIME_END = 0
        EXPECTED_LONGEST_ITEMS_SEQUENCE = 0
        EXPECTED_NUMBER_OF_SESSIONS = 0

        self.assertEqual(EXPECTED_MAP, sessions.session_items_actions_map)
        self.assertEqual(EXPECTED_TIME_START, sessions.time_start)
        self.assertEqual(EXPECTED_TIME_END, sessions.time_end)
        self.assertEqual(EXPECTED_LONGEST_ITEMS_SEQUENCE, sessions.longest_items_vector_size)
        self.assertEqual(EXPECTED_NUMBER_OF_SESSIONS, sessions.number_of_sessions)

    def test_append(self):

        expected_number_of_sessions = 1
        expected_longest_items_seq = 1
        expected_ts = 1
        expected_tend = 1
        session_id = '0'

        sessions = Sessions(event_session_key=SESSION_KEY,
                            event_product_key=ITEM_KEY,
                            event_time_key=T_KEY,
                            event_action_key=ACTION_KEY,
                            event_action_weights=ACTION_WEIGHTS)
        sessions.append(EVENT1)

        self.assertEqual(expected_ts, sessions.time_start)
        self.assertEqual(expected_tend, sessions.time_end)
        self.assertEqual(expected_longest_items_seq, sessions.longest_items_vector_size)
        self.assertEqual(expected_number_of_sessions, sessions.number_of_sessions)
        self.assertIn(session_id, sessions.session_items_actions_map)

    def test_save_and_load(self):

        EMPTY_SESSION_PATH = 'tdata/test_save_empty_session.pkl'
        SESSION_PATH = 'tdata/test_save_session.pkl'

        sessions = Sessions(event_session_key=SESSION_KEY,
                            event_product_key=ITEM_KEY,
                            event_time_key=T_KEY,
                            event_action_key=ACTION_KEY,
                            event_action_weights=ACTION_WEIGHTS)
        # Save empty
        sessions.save(EMPTY_SESSION_PATH)
        # Append event
        sessions.append(EVENT1)
        # Save
        sessions.save(SESSION_PATH)

        del sessions

        # Load empty
        empty_sessions = Sessions(event_session_key=SESSION_KEY,
                                  event_product_key=ITEM_KEY,
                                  event_time_key=T_KEY,
                                  event_action_key=ACTION_KEY,
                                  event_action_weights=ACTION_WEIGHTS)
        empty_sessions.load(EMPTY_SESSION_PATH)

        # Load processed
        sessions = Sessions(event_session_key=SESSION_KEY,
                            event_product_key=ITEM_KEY,
                            event_time_key=T_KEY,
                            event_action_key=ACTION_KEY,
                            event_action_weights=ACTION_WEIGHTS)
        sessions.load(SESSION_PATH)

        # TESTS - EMPTY
        empty_expected_map = dict()
        empty_expected_ts = 1_000_000_000_000_000
        empty_expected_tend = 0
        empty_expected_seq = 0
        empty_expected_no_sessions = 0
        self.assertEqual(empty_expected_map, empty_sessions.session_items_actions_map)
        self.assertEqual(empty_expected_ts,empty_sessions.time_start)
        self.assertEqual(empty_expected_tend, empty_sessions.time_end)
        self.assertEqual(empty_expected_seq, empty_sessions.longest_items_vector_size)
        self.assertEqual(empty_expected_no_sessions, empty_sessions.number_of_sessions)

        # TESTS - processed
        no_sessions = 1
        longest_seq = 1
        ts = 1
        tend = 1
        session_id = '0'
        self.assertEqual(ts, sessions.time_start)
        self.assertEqual(tend, sessions.time_end)
        self.assertEqual(longest_seq, sessions.longest_items_vector_size)
        self.assertEqual(no_sessions, sessions.number_of_sessions)
        self.assertIn(session_id, sessions.session_items_actions_map)

    def test_add(self):
        sessions1 = Sessions(event_session_key=SESSION_KEY,
                            event_product_key=ITEM_KEY,
                            event_time_key=T_KEY,
                            event_action_key=ACTION_KEY,
                            event_action_weights=ACTION_WEIGHTS)
        sessions2 = Sessions(event_session_key=SESSION_KEY,
                            event_product_key=ITEM_KEY,
                            event_time_key=T_KEY,
                            event_action_key=ACTION_KEY,
                            event_action_weights=ACTION_WEIGHTS)
        sessions3 = Sessions(event_session_key=SESSION_KEY,
                            event_product_key=ITEM_KEY,
                            event_time_key=T_KEY,
                            event_action_key=ACTION_KEY,
                            event_action_weights=ACTION_WEIGHTS)
        sessions4 = Sessions(event_session_key=SESSION_KEY,
                             event_product_key=ITEM_KEY,
                             event_time_key=T_KEY,
                             event_action_key=ACTION_KEY,
                             event_action_weights=ACTION_WEIGHTS)
        sessions1.append(EVENT1)
        sessions2.append(EVENT2)
        sessions3.append(EVENT3)

        sessions12 = sessions1 + sessions2

        sessions123 = sessions12 + sessions3
        sessions123.update_weights_of_purchase_session(EVENT4['sid'],
                                                       additive_factor=1,
                                                       bought_products=EVENT_4_PRODUCTS)

        weights12 = sessions12.session_items_actions_map['0'][-1]
        weights123 = sessions123.session_items_actions_map['0'][-1]

        w12 = sum(weights12)
        w123 = sum(weights123)
        diff = w123 - w12
        expected_diff = 2

        self.assertEqual(expected_diff, diff)

    def test_no_actions(self):
        sessions = Sessions(event_session_key=SESSION_KEY,
                            event_product_key=ITEM_KEY,
                            event_time_key=T_KEY,
                            event_action_key=None,
                            event_action_weights=None)

        self.assertIsInstance(sessions, Sessions)

        EXPECTED_MAP = dict()
        EXPECTED_TIME_START = 1_000_000_000_000_000
        EXPECTED_TIME_END = 0
        EXPECTED_LONGEST_ITEMS_SEQUENCE = 0
        EXPECTED_NUMBER_OF_SESSIONS = 0

        self.assertEqual(EXPECTED_MAP, sessions.session_items_actions_map)
        self.assertEqual(EXPECTED_TIME_START, sessions.time_start)
        self.assertEqual(EXPECTED_TIME_END, sessions.time_end)
        self.assertEqual(EXPECTED_LONGEST_ITEMS_SEQUENCE, sessions.longest_items_vector_size)
        self.assertEqual(EXPECTED_NUMBER_OF_SESSIONS, sessions.number_of_sessions)


if __name__ == '__main__':
    unittest.main()



from datetime import datetime
from typing import Any, List


def clean_time(times: Any,
               time_to_numeric: bool,
               time_to_datetime: bool,
               datetime_format: str) -> List:
    """
    Function cleans timestamps if they don't have valid data type.

    Parameters
    ----------
    times : List
        The list with timestamps.

    time_to_numeric : bool
        Should time be transformed to numeric?

    time_to_numeric : bool, default = True
        Transforms input timestamps to float values.

    time_to_datetime : bool, default = False
        Transforms input timestamps to datatime objects. Setting `datetime_format` parameter is required.

    datetime_format : str
        The format of datetime object.

    Returns
    -------
    : List
        Parsed times.
    """

    parsed_times = times

    if isinstance(times, List):
        if time_to_numeric:
            parsed_times = [float(x) for x in times]

        if time_to_datetime:
            parsed_times = [datetime.strptime(x, datetime_format) for x in times]
    else:
        if time_to_numeric:
            parsed_times = float(times)

        if time_to_datetime:
            parsed_times = datetime.strptime(times, datetime_format)

    return parsed_times



import pickle
from wsknn.preprocessing.utils.calc import get_larger_value, get_smaller_value
from wsknn.preprocessing.utils.transform import merge_dicts, parse_seconds_to_dt


class Items:
    """
    Class stores item-sessions map and its basic properties. The core object is a dictionary of unique items (keys)
    that are pointing to the specific sessions and their timestamps (lists).

    Parameters
    ----------
    event_session_key
        The name of a session key.

    event_product_key
        The name of an item key.

    event_time_key
        The name of a timestamp key.

    Attributes
    ----------
    item_sessions_map : Dict
        The item-sessions mapper ``{item_id: [[sessions], [first timestamp of each session]]}``.

    time_start : int, default = 1_000_000_000_000_000
        The initial timestamp, first event in whole dataset.

    time_end : int, default = 0
        The timestamp of the last event in dataset.

    longest_sessions_vector_size: int, default = 0
        The longest sequence of sessions that contained the item.

    number_of_items : int, default = 0
        The number of items within the ``item_sessions_map`` object.

    event_session_key
        See the ``event_session_key`` parameter.

    event_product_key
        See the ``event_product_key`` parameter.

    event_time_key
        See the ``event_time_key`` parameter.

    metadata : str
        A description of the ``Items`` class.

    Methods
    -------
    append(event)
        Appends a single event to the item-sessions map.

    export(filename)
        Method exports created mapping to a pickled dictionary.

    load(filename)
        Loads pickled ``Items`` object into a new instance of a class.

    save_object(filename)
        Items object is stored as a Python pickle binary file.

    __add__(Users)
        Adds other Items object. It is a set operation. Therefore, sessions that are assigned to the same
        item within Items(1) and Items(2) won't be duplicated.

    __str__()
        The basic info about the class.

    Raises
    ------
    TypeError
        Timestamps are not datatime objects or numerical objects.

    """

    def __init__(self,
                 event_session_key: str,
                 event_product_key: str,
                 event_time_key: str):

        self.item_sessions_map = dict()
        self.time_start = 1_000_000_000_000_000
        self.time_end = 0
        self.longest_sessions_vector_size = 0
        self.number_of_items = 0

        self.event_session_key = event_session_key
        self.event_product_key = event_product_key
        self.event_time_key = event_time_key

        self.metadata = self._get_metadata()

    @staticmethod
    def _get_metadata():
        meta = """
        The items object is a dictionary of unique items (keys) that are pointing to the specific sessions and their 
        timestamps (lists).

        Key map:

        item_sessions_map = {
            item_id: (
                [sequence_of_sessions],
                [sequence_of_the_first_session_timestamps]
            )
        }

        Other keys:
        - time_start: the first date in a dataset.
        - time_end: the last date in a dataset.
        - longest_sessions_vector_size: size of the longest sequence sessions,
        - number_of_items: the length of item_sessions_map.
        """
        return meta

    def _update_number_of_items(self):
        self.number_of_items = self.number_of_items + 1

    def _update_longest_session_size(self, session_length: int):
        if session_length > self.longest_sessions_vector_size:
            self.longest_sessions_vector_size = session_length

    def _update_first_timestamp(self, ts):
        self.time_start = ts

    def _update_last_timestamp(self, ts):
        self.time_end = ts

    def _update_timestamps(self, ts: int):

        # First
        try:
            if ts < self.time_start:
                self._update_first_timestamp(ts)

            # Last
            if ts > self.time_end:
                self._update_last_timestamp(ts)
        except TypeError:
            raise TypeError('Timestamp values should have datetime or numeric types to allow comparisons!')

    def _append_item_session_and_timestamp(self, item: str, session: str, timestamp: int):
        """
        Function appends item session and timestamp to existing list of sessions and timestamps.

        Parameters
        ----------
        item : str
            ID of the product.

        session : str
            ID of the session.

        timestamp : int
            Timestamp of the event.
        """
        if session in self.item_sessions_map[item][0]:

            # Session exists, upload timestamp if needed
            idx = self.item_sessions_map[item][0].index(session)

            past_ts = self.item_sessions_map[item][1][idx]

            if timestamp < past_ts:
                self.item_sessions_map[item][1][idx] = timestamp

            # Update the first and the last time reading if needed
            self._update_timestamps(timestamp)
        else:
            # Append session to sessions
            self.item_sessions_map[item][0].append(session)
            # Append timestamp to timestamps
            self.item_sessions_map[item][1].append(timestamp)
            # Update the first and the last time reading if needed
            self._update_timestamps(timestamp)
            # Update longest session info
            self._update_longest_session_size(len(self.item_sessions_map[item][0]))

    def append(self, event: dict):
        """
        Method appends given event into internal structure of the class.

        Parameters
        ----------
        event : Dict
            Expected keys: event time, event session, event product
        """

        item = event[self.event_product_key]
        session = event[self.event_session_key]
        dt = event[self.event_time_key]

        if item in self.item_sessions_map:
            self._append_item_session_and_timestamp(item, session, dt)
        else:
            # Create Item-sessions map
            self.item_sessions_map[item] = (
                [session],
                [dt]
            )
            # Update number of items in the dictionary
            self._update_number_of_items()
            # Update the first and the last time reading if needed
            self._update_timestamps(dt)
            # Update longest session info
            self._update_longest_session_size(len(self.item_sessions_map[item][0]))

    def export_to_dict(self, filename: str):
        """
        Method exports object's attributes as a dictionary in a pickled object.

        Parameters
        ----------
        filename : str
            The path to the pickled object. If suffix .pkl is not given then methods appends it into the file.

        Notes
        -----
        Dictionary keys are:
            * ``map``: item-sessions map,
            * ``time_start``: the earliest event time,
            * ``time_end``: the latest event time,
            * ``longest_sessions_vector_size``: the number of sessions in the longest item-sequence,
            * ``number_of_items``: total number of parsed items,
            * ``metadata``: additional class info.
        """
        pkl = '.pkl'
        if not filename.endswith(pkl):
            filename = filename + pkl

        objects_att_dict = {
            'map': self.item_sessions_map,
            'time_start': self.time_start,
            'time_end': self.time_end,
            'longest_sessions_vector_size': self.longest_sessions_vector_size,
            'number_of_items': self.number_of_items,
            'metadata': self.metadata
        }

        with open(filename, 'wb') as storage:
            pickle.dump(objects_att_dict, storage)

    def save(self, filename: str):
        """
        Method saves ``Items`` class object into a pickled file.

        Parameters
        ----------
        filename : str
            The path to the pickled object. If suffix .pkl is not given then methods appends it into the file.
        """

        pkl = '.pkl'
        if not filename.endswith(pkl):
            filename = filename + pkl

        with open(filename, 'wb') as storage:
            pickle.dump(self, storage)

    def load(self, filename: str):
        """
        Method loads pickled ``Items`` class object and assigns its properties and data into the class instance.

        Parameters
        ----------
        filename : str
            The path to the file.

        Raises
        ------
        IOError
            Cannot load data into ``Items`` object with parsed events.
        """
        # Check if current object has any items to avoid overwriting
        if self.item_sessions_map:
            raise IOError('Cannot load data into a populated object! '
                          'Create empty object to load data from pickled file.')

        # Load
        with open(filename, 'rb') as stored_data:
            self.__dict__.update(pickle.load(stored_data).__dict__)

    def __add__(self, other):
        """
        Adds items and mapped sessions from one ``Items`` object into another.

        Parameters
        ----------
        other : Items

        Returns
        -------
        merged : Items
            The merged and updated ``Items`` object.
        """
        merged = Items(other.event_session_key, other.event_product_key, other.event_time_key)
        merged.item_sessions_map = merge_dicts(self.item_sessions_map, other.item_sessions_map)
        merged.time_start = get_smaller_value(self.time_start, other.time_start)
        merged.time_end = get_larger_value(self.time_end, other.time_end)
        merged.longest_sessions_vector_size = get_larger_value(self.longest_sessions_vector_size,
                                                               other.longest_sessions_vector_size)
        merged.number_of_items = len(merged.item_sessions_map)
        return merged

    def __str__(self):
        if self.item_sessions_map:
            try:
                ts = parse_seconds_to_dt(self.time_start)
                te = parse_seconds_to_dt(self.time_end)
            except ValueError:
                # TODO
                ts = self.time_start
                te = self.time_end
            msg = f'Items object statistics:\n' \
                  f'*Number of unique items: {self.number_of_items}\n' \
                  f'*The longest sessions vector size: {self.longest_sessions_vector_size}\n' \
                  f'*Period start: {ts}\n' \
                  f'*Period end: {te}'
            return msg
        else:
            msg = 'Empty object. Append events or load pickled item-sessions map!'
            return msg



from wsknn.fit_transform import fit, Items, Sessions
from wsknn.predict import predict
from wsknn.model.wsknn import WSKNN
from wsknn.preprocessing.parse_static import parse_files
from wsknn.preprocessing.static_parsers.pandas_parser import parse_pandas


__version__ = '1.2.0'



import unittest
from wsknn.preprocessing.structure.item import Items


EVENT_PRODUCT_KEY = 'product'
EVENT_SESSION_KEY = 'session'
EVENT_TIME_KEY = 'time'

class TestItemsClass(unittest.TestCase):

    def test_init(self):
        items = Items(event_product_key=EVENT_PRODUCT_KEY,
                      event_session_key=EVENT_SESSION_KEY,
                      event_time_key=EVENT_TIME_KEY)
        self.assertIsInstance(items, Items)

        EXPECTED_MAP = dict()
        EXPECTED_TIME_START = 1_000_000_000_000_000
        EXPECTED_TIME_END = 0
        EXPECTED_LONGEST_SESSIONS_VECTOR_SIZE = 0
        EXPECTED_NUMBER_OF_ITEMS = 0

        self.assertEqual(EXPECTED_MAP, items.item_sessions_map)
        self.assertEqual(EXPECTED_TIME_START, items.time_start)
        self.assertEqual(EXPECTED_TIME_END, items.time_end)
        self.assertEqual(EXPECTED_LONGEST_SESSIONS_VECTOR_SIZE, items.longest_sessions_vector_size)
        self.assertEqual(EXPECTED_NUMBER_OF_ITEMS, items.number_of_items)

    def test_append(self):
        EVENT = {
            EVENT_PRODUCT_KEY: 'test_product',
            EVENT_SESSION_KEY: 'test_session',
            EVENT_TIME_KEY: 0
        }
        EXPECTED_NUMBER_OF_ITEMS = 1
        EXPECTED_LONGEST_SESSION_SIZE = 1
        EXPECTED_TIME_START = 0
        EXPECTED_TIME_END = 0
        PRODUCT = 'test_product'

        items = Items(event_product_key=EVENT_PRODUCT_KEY,
                      event_session_key=EVENT_SESSION_KEY,
                      event_time_key=EVENT_TIME_KEY)
        items.append(EVENT)

        self.assertEqual(EXPECTED_TIME_START, items.time_start)
        self.assertEqual(EXPECTED_TIME_END, items.time_end)
        self.assertEqual(EXPECTED_LONGEST_SESSION_SIZE, items.longest_sessions_vector_size)
        self.assertEqual(EXPECTED_NUMBER_OF_ITEMS, items.number_of_items)
        self.assertIn(PRODUCT, items.item_sessions_map)

    def test_save_and_load(self):
        EVENT = {
            EVENT_PRODUCT_KEY: 'test_product',
            EVENT_SESSION_KEY: 'test_session',
            EVENT_TIME_KEY: 0
        }

        EMPTY_ITEM_PATH = 'tdata/test_save_empty_item.pkl'
        ITEM_PATH = 'tdata/test_save_item.pkl'

        items = Items(event_product_key=EVENT_PRODUCT_KEY,
                      event_session_key=EVENT_SESSION_KEY,
                      event_time_key=EVENT_TIME_KEY)

        # Save empty
        items.save(EMPTY_ITEM_PATH)
        # Append event
        items.append(EVENT)
        # Save
        items.save(ITEM_PATH)

        del items

        # Load empty
        empty_items = Items(event_product_key=EVENT_PRODUCT_KEY,
                            event_session_key=EVENT_SESSION_KEY,
                            event_time_key=EVENT_TIME_KEY)
        empty_items.load(EMPTY_ITEM_PATH)

        # Load processed
        items = Items(event_product_key=EVENT_PRODUCT_KEY,
                      event_session_key=EVENT_SESSION_KEY,
                      event_time_key=EVENT_TIME_KEY)
        items.load(ITEM_PATH)

        # TESTS - EMPTY
        EMPTY_EXPECTED_MAP = dict()
        EMPTY_EXPECTED_TIME_START = 1_000_000_000_000_000
        EMPTY_EXPECTED_TIME_END = 0
        EMPTY_EXPECTED_LONGEST_SESSIONS_VECTOR_SIZE = 0
        EMPTY_EXPECTED_NUMBER_OF_ITEMS = 0
        self.assertEqual(EMPTY_EXPECTED_MAP, empty_items.item_sessions_map)
        self.assertEqual(EMPTY_EXPECTED_TIME_START, empty_items.time_start)
        self.assertEqual(EMPTY_EXPECTED_TIME_END, empty_items.time_end)
        self.assertEqual(EMPTY_EXPECTED_LONGEST_SESSIONS_VECTOR_SIZE, empty_items.longest_sessions_vector_size)
        self.assertEqual(EMPTY_EXPECTED_NUMBER_OF_ITEMS, empty_items.number_of_items)

        # TESTS - processed
        EXPECTED_NUMBER_OF_ITEMS = 1
        EXPECTED_LONGEST_SESSION_SIZE = 1
        EXPECTED_TIME_START = 0
        EXPECTED_TIME_END = 0
        PRODUCT = 'test_product'
        self.assertEqual(EXPECTED_TIME_START, items.time_start)
        self.assertEqual(EXPECTED_TIME_END, items.time_end)
        self.assertEqual(EXPECTED_LONGEST_SESSION_SIZE, items.longest_sessions_vector_size)
        self.assertEqual(EXPECTED_NUMBER_OF_ITEMS, items.number_of_items)
        self.assertIn(PRODUCT, items.item_sessions_map)

    def test_add(self):
        EVENT_0 = {
            EVENT_PRODUCT_KEY: 'test_product',
            EVENT_SESSION_KEY: 'test_session',
            EVENT_TIME_KEY: 0
        }
        EVENT_1 = {
            EVENT_PRODUCT_KEY: 'other_test_product',
            EVENT_SESSION_KEY: 'test_session',
            EVENT_TIME_KEY: 1000
        }
        EVENT_2 = {
            EVENT_PRODUCT_KEY: 'test_product',
            EVENT_SESSION_KEY: 'other_test_session',
            EVENT_TIME_KEY: 500
        }

        items_0 = Items(event_product_key=EVENT_PRODUCT_KEY,
                        event_session_key=EVENT_SESSION_KEY,
                        event_time_key=EVENT_TIME_KEY)
        items_0.append(EVENT_0)
        items_1 = Items(event_product_key=EVENT_PRODUCT_KEY,
                        event_session_key=EVENT_SESSION_KEY,
                        event_time_key=EVENT_TIME_KEY)
        items_1.append(EVENT_1)
        items_2 = Items(event_product_key=EVENT_PRODUCT_KEY,
                        event_session_key=EVENT_SESSION_KEY,
                        event_time_key=EVENT_TIME_KEY)
        items_2.append(EVENT_2)

        # keys: 'test_product', 'other_test_product'
        items_01 = items_0 + items_1

        # keys: 'other_test_product', 'test_product'
        # time_start = 500
        # time_end = 1000
        items_12 = items_1 + items_2

        # keys: 'test_product', 'other_test_product'
        # time_start = 0
        # time_end = 1000
        items_012 = items_0 + items_1 + items_2

        # keys: 'test_product'
        # time_start = 0
        # time_end = 0
        items_00 = items_0 + items_0

        EXPECTED_KEYS_01 = ['test_product', 'other_test_product']
        EXPECTED_KEYS_12 = EXPECTED_KEYS_01
        EXPECTED_KEYS_012 = EXPECTED_KEYS_01
        EXPECTED_KEYS_00 = ['test_product']
        EXPECTED_TIME_START_01 = 0
        EXPECTED_TIME_END_01 = 1000
        EXPECTED_TIME_START_12 = 500
        EXPECTED_TIME_END_12 = 1000
        EXPECTED_TIME_START_012 = 0
        EXPECTED_TIME_END_012 = 1000
        EXPECTED_TIME_START_00 = 0
        EXPECTED_TIME_END_00 = 0

        for k in items_01.item_sessions_map.keys():
            self.assertIn(k, EXPECTED_KEYS_01)
        for k in items_12.item_sessions_map.keys():
            self.assertIn(k, EXPECTED_KEYS_12)
        for k in items_012.item_sessions_map.keys():
            self.assertIn(k, EXPECTED_KEYS_012)
        for k in items_00.item_sessions_map.keys():
            self.assertIn(k, EXPECTED_KEYS_00)

        self.assertEqual(items_01.time_start, EXPECTED_TIME_START_01)
        self.assertEqual(items_01.time_end, EXPECTED_TIME_END_01)
        self.assertEqual(items_12.time_start, EXPECTED_TIME_START_12)
        self.assertEqual(items_12.time_end, EXPECTED_TIME_END_12)
        self.assertEqual(items_012.time_start, EXPECTED_TIME_START_012)
        self.assertEqual(items_012.time_end, EXPECTED_TIME_END_012)
        self.assertEqual(items_00.time_start, EXPECTED_TIME_START_00)
        self.assertEqual(items_00.time_end, EXPECTED_TIME_END_00)

    def test_str(self):
        EVENT = {
            EVENT_PRODUCT_KEY: 'test_product',
            EVENT_SESSION_KEY: 'test_session',
            EVENT_TIME_KEY: 0
        }

        empty_items = Items(event_product_key=EVENT_PRODUCT_KEY,
                            event_session_key=EVENT_SESSION_KEY,
                            event_time_key=EVENT_TIME_KEY)
        items = Items(event_product_key=EVENT_PRODUCT_KEY,
                      event_session_key=EVENT_SESSION_KEY,
                      event_time_key=EVENT_TIME_KEY)
        items.append(EVENT)

        EMPTY_MSG = 'Empty object. Append events or load pickled item-sessions map!'
        self.assertEqual(empty_items.__str__(), EMPTY_MSG)


if __name__ == '__main__':
    unittest.main()



from typing import Dict, Iterable, IO, Tuple, List
import pandas as pd
from more_itertools import locate
from tqdm import tqdm
from wsknn.preprocessing.static_parsers.checkers.validation import check_event_keys_and_values, is_user_item_interaction
from wsknn.preprocessing.static_parsers.cleaners.time_transform import clean_time
from wsknn.preprocessing.structure.item import Items
from wsknn.preprocessing.structure.session import Sessions


def parse_fn(dataset: Iterable,
             allowed_actions: Dict,
             purchase_action_name: str,
             session_id_key: str,
             product_key: str,
             action_key: str,
             time_key: str,
             time_to_numeric: bool,
             time_to_datetime: bool,
             datetime_format: str,
             progress_bar: bool) -> (Items, Sessions):
    """
    Function parses given dataset into Sessions and Items objects.

    Parameters
    ----------
    dataset : Iterable
        Object with events.

    allowed_actions : Dict, optional
        Allowed actions and their weights.

    purchase_action_name: Any, optional
        The name of the final action (it is required to apply weight into the session vector).

    session_id_key : str
        The name of the session key.

    product_key : str
        The name of the product key.

    action_key : str
        The name of the event action type key.

    time_key : str
        The name of the event timestamp key.

    time_to_numeric : bool, default = True
        Transforms input timestamps to float values.

    time_to_datetime : bool, default = False
        Transforms input timestamps to datatime objects. Setting `datetime_format` parameter is required.

    datetime_format : str
        The format of datetime object.

    progress_bar : bool
        Show parsing progress.

    Returns
    -------
    ItemsMap, SessionsMap : Items, Sessions
    """

    # Initialize Items and Sessions

    items_obj = Items(event_session_key=session_id_key,
                      event_product_key=product_key,
                      event_time_key=time_key)

    sessions_obj = Sessions(event_session_key=session_id_key,
                            event_product_key=product_key,
                            event_time_key=time_key,
                            event_action_key=action_key,
                            event_action_weights=allowed_actions)

    possible_actions_list = []
    if allowed_actions is not None:
        possible_actions_list = list(allowed_actions.keys())

    for event in (tqdm(dataset, disable=(not progress_bar))):
        event = check_event_keys_and_values(event,
                                            session_id_key,
                                            product_key,
                                            time_key,
                                            action_key)
        # Check if params are returned
        if event:
            # parse times
            if time_to_numeric or time_to_datetime:
                event[time_key] = clean_time(times=event[time_key],
                                             time_to_numeric=time_to_numeric,
                                             time_to_datetime=time_to_datetime,
                                             datetime_format=datetime_format)

            action = event.get(action_key, False)
            if not action:
                items_obj.append(event)
                sessions_obj.append(event)
            else:
                if action != purchase_action_name:
                    # Is session user interaction?
                    if is_user_item_interaction(action, possible_actions_list):
                        # Append Event to Items and Sessions
                        items_obj.append(event)
                        sessions_obj.append(event)
                else:
                    # It is a purchase, update weights accordingly
                    purchase_additive_factor = allowed_actions[purchase_action_name]
                    sessions_obj.update_weights_of_purchase_session(event[session_id_key], purchase_additive_factor)

    return items_obj, sessions_obj


def parse_stream(events: IO,
                 sep: str,
                 allowed_actions: Dict,
                 purchase_action_name: str,
                 session_index: int,
                 product_index: int,
                 time_index: int,
                 action_index: int,
                 time_to_numeric: bool,
                 time_to_datetime: bool,
                 datetime_format: str,
                 ignore_errors: bool = True,
                 header_names: Dict = None,
                 progress_bar: bool = False):
    """
    Function parses given stream of values.

    Parameters
    ----------
    events : IO
        Stream to file.

    sep : str
        Separator between file stream records.

    allowed_actions : Dict, optional
        Allowed actions and their weights.

    purchase_action_name: Any, optional
        The name of the final action (it is required to apply weight into the session vector).

    session_index : int
        The index of the session.

    product_index : int
        The index of the product.

    action_index : int
        The index of the event action.

    time_index : int
        The index of the event timestamp.

    time_to_numeric : bool, default = False
        Transforms input timestamps to float values.

    time_to_datetime : bool, default = False
        Transforms input timestamps to datatime objects. Setting ``datetime_format`` parameter is required.

    datetime_format : str
        The format of datetime object.

    ignore_errors : bool, default=True
        Ignore rows that raise exceptions.

    header_names : List, default = None
        Key names applied to the data.

    progress_bar : bool, default = False
        Show parsing progress.

    Returns
    -------
    items_obj, sessions_obj : Items, Sessions
    """

    # Initialize Items and Sessions

    if header_names is None:
        if action_index is None:
            header_names = {
                session_index: 'session',
                product_index: 'item',
                time_index: 'ts'
            }
        else:
            header_names = {
                session_index: 'session',
                product_index: 'item',
                time_index: 'ts',
                action_index: 'action'
            }

    items_obj = Items(event_session_key=header_names[session_index],
                      event_product_key=header_names[product_index],
                      event_time_key=header_names[time_index])

    if action_index is not None:
        sessions_obj = Sessions(event_session_key=header_names[session_index],
                                event_product_key=header_names[product_index],
                                event_time_key=header_names[time_index],
                                event_action_key=header_names[action_index],
                                event_action_weights=allowed_actions)
    else:
        sessions_obj = Sessions(event_session_key=header_names[session_index],
                                event_product_key=header_names[product_index],
                                event_time_key=header_names[time_index],
                                event_action_key=None,
                                event_action_weights=allowed_actions)

    if allowed_actions is None:
        possible_actions_list = None
    else:
        possible_actions_list = list(allowed_actions.keys())

    for raw_event in (tqdm(events, disable=(not progress_bar))):

        try:
            splitted = raw_event.split(sep)
            splitted = [s.strip() for s in splitted]
        except Exception as ex:
            if ignore_errors:
                continue
            else:
                raise ex

        if action_index is not None:
            event = {
                header_names[session_index]: splitted[session_index],
                header_names[product_index]: splitted[product_index],
                header_names[action_index]: splitted[action_index],
                header_names[time_index]: splitted[time_index]
            }
        else:
            event = {
                header_names[session_index]: splitted[session_index],
                header_names[product_index]: splitted[product_index],
                header_names[time_index]: splitted[time_index]
            }

        event = check_event_keys_and_values(event,
                                            header_names[session_index],
                                            header_names[product_index],
                                            header_names[time_index],
                                            header_names.get(action_index, None))
        # Check if params are returned
        if event:
            # parse times
            if time_to_numeric or time_to_datetime:
                event[header_names[time_index]] = clean_time(times=event[header_names[time_index]],
                                                             time_to_numeric=time_to_numeric,
                                                             time_to_datetime=time_to_datetime,
                                                             datetime_format=datetime_format)

            action = event.get(action_index, False)
            if not action:
                items_obj.append(event)
                sessions_obj.append(event)
            else:
                if action != purchase_action_name:
                    # Is session user interaction?
                    if is_user_item_interaction(action, possible_actions_list):
                        # Append Event to Items and Sessions
                        items_obj.append(event)
                        sessions_obj.append(event)
                else:
                    # It is a purchase, update weights accordingly
                    purchase_additive_factor = allowed_actions[purchase_action_name]
                    sessions_obj.update_weights_of_purchase_session(event[header_names[session_index]],
                                                                    purchase_additive_factor)

    return items_obj, sessions_obj



import pickle
from typing import Union, Dict, Iterable

from wsknn.preprocessing.utils.calc import get_larger_value, get_smaller_value
from wsknn.preprocessing.utils.transform import merge_dicts, parse_seconds_to_dt


class Sessions:
    """
    Class stores session-items mapping and its basic properties. The core object is a dictionary of unique
    sessions (keys) that are pointing to the session-items and their timestamps (lists).

    Parameters
    ----------
    event_session_key
        The name of a session key.

    event_product_key
        The name of an item key.

    event_time_key
        The name of a timestamp key.

    event_action_key : str, optional
        The name of a key with actions that may be used for neighbors filtering.

    event_action_weights : Dict, optional
        Dictionary with weights that should be assigned to event actions.

    Attributes
    ----------
    session_items_actions_map : Dict
        The session-items mapper ``{session_id: [[items], [timestamps], [optional - actions], [optional - weights]]}``.

    time_start : int, default = 1_000_000_000_000_000
        The initial timestamp, first event in whole dataset.

    time_end : int, default = 0
        The timestamp of the last event in dataset.

    longest_items_vector_size: int, default = 0
        The longest sequence of products.

    number_of_sessions : int, default = 0
        The number of sessions within a ``session_items_actions_map`` object.

    event_session_key
        See the ``event_session_key`` parameter.

    event_product_key
        See the ``event_product_key`` parameter.

    event_time_key
        See the ``event_time_key`` parameter.

    event_action_key
        See the ``event_action_key`` parameter.

    action_weights
        See the ``event_action_weights`` parameter.

    metadata : str
        A description of the ``Sessions`` class.

    Methods
    -------
    append(event)
        Appends a single event to the session-items map.

    export(filename)
        Method exports created mapping to pickled dictionary.

    load(filename)
        Loads pickled ``Sessions`` object into a new instance of a class.

    save_object(filename)
        Users object is stored as a Python pickle binary file.

    update_weights_of_purchase_session()
        Adds constant value to all weights if session ended up with purchase.

    __add__(Users)
        Adds other Sessions object.

    __str__()
        The basic info about the class.

    """

    def __init__(self,
                 event_session_key: str,
                 event_product_key: str,
                 event_time_key: str,
                 event_action_key: str = None,
                 event_action_weights: Dict = None):

        self.session_items_actions_map = dict()
        self.time_start = 1_000_000_000_000_000
        self.time_end = 0
        self.longest_items_vector_size = 0
        self.number_of_sessions = 0

        self.event_session_key = event_session_key
        self.event_product_key = event_product_key
        self.event_time_key = event_time_key
        self.event_action_key = event_action_key
        self.action_weights = event_action_weights

        self.metadata = self._get_metadata()

    @staticmethod
    def _get_metadata():
        meta = """
        The sessions object is a dictionary of unique sessions (keys) that are pointing to the specific items and their 
        timestamps and, optionally, weighting factors.

        Key map:

        map = {
            session_id: [
                [items],
                [timestamps],
                [optional - actions],
                [optional - action weights]
            ]
        }

        Other keys:
        - time_start: the first date in a dataset.
        - time_end: the last date in a dataset.
        - longest_sessions_vector_size: size of the longest sequence sessions,
        - number_of_items: the length of item_sessions_map.
        """
        return meta

    def _update_number_of_sessions(self):
        self.number_of_sessions = self.number_of_sessions + 1

    def _update_longest_event_sequence_size(self, seq_length: int):
        if seq_length > self.longest_items_vector_size:
            self.longest_items_vector_size = seq_length

    def _update_first_timestamp(self, ts):
        self.time_start = ts

    def _update_last_timestamp(self, ts):
        self.time_end = ts

    def _update_timestamps(self, ts: int):

        # First
        if ts < self.time_start:
            self._update_first_timestamp(ts)

        # Last
        if ts > self.time_end:
            self._update_last_timestamp(ts)

    def _append_session_item_and_timestamp(self,
                                           session: str,
                                           item: str,
                                           timestamp: int,
                                           action: Union[str, None] = None):
        """
        Function appends item session and timestamp to existing list of sessions and timestamps.
        """
        # Append item
        self.session_items_actions_map[session][0].append(item)
        # Append timestamp
        self.session_items_actions_map[session][1].append(timestamp)

        if action is not None:
            # Append action type
            self.session_items_actions_map[session][2].append(action)

        if self.action_weights is not None:
            self.session_items_actions_map[session][3].append(self.action_weights.get(action))

        # Update the first and the last time reading if needed
        self._update_timestamps(timestamp)
        # Update longest session info
        self._update_longest_event_sequence_size(len(self.session_items_actions_map[session][0]))

    def append(self, event: dict):
        """
        Method appends the event into session-items map.

        Parameters
        ----------
        event : Dict
            Single event that consists: session id, timestamp, product, and optionally action type.
        """

        item = event[self.event_product_key]
        session = event[self.event_session_key]
        dt = event[self.event_time_key]

        if self.event_action_key is not None:
            action = event[self.event_action_key]
        else:
            action = None

        if session in self.session_items_actions_map:
            self._append_session_item_and_timestamp(session, item, dt, action)
        else:
            # Create Item-sessions map

            if action is None:
                self.session_items_actions_map[session] = (
                    [item],
                    [dt]
                )
            else:
                if self.action_weights is None:
                    self.session_items_actions_map[session] = (
                        [item],
                        [dt],
                        [action]
                    )
                else:
                    self.session_items_actions_map[session] = (
                        [item],
                        [dt],
                        [action],
                        [self.action_weights.get(action, 0.001)]
                    )
            # Update number of sessions in the dictionary
            self._update_number_of_sessions()
            # Update the first and the last time reading if needed
            self._update_timestamps(dt)
            # Update longest session info
            self._update_longest_event_sequence_size(len(self.session_items_actions_map[session][0]))

    def update_weights_of_purchase_session(self, session_id,
                                           additive_factor=0.0,
                                           bought_products: Iterable = None):
        """
        Method updates item weights accordingly to the given additive factor. Running this method is required when
        we generate optional weights list.

        Parameters
        ----------
        session_id : Any
            The index of a session.

        additive_factor : float
            Weight added to the item.

        bought_products : List, optional
            If provided, then only those items weights are updated.
        """
        for idx, item in enumerate(self.session_items_actions_map[session_id][0]):
            old_value = self.session_items_actions_map[session_id][3][idx]
            new_value = old_value + additive_factor
            if bought_products is not None:
                if item in bought_products:
                    self.session_items_actions_map[session_id][3][idx] = new_value
            else:
                self.session_items_actions_map[session_id][3][idx] = new_value

    def export_to_dict(self, filename: str):
        """
        Method exports object's attributes as a dictionary in a pickled object.

        Parameters
        ----------
        filename : str
            The path to the pickled object. If suffix .pkl is not given then methods appends it into the file.

        Notes
        -----
        Dictionary keys are:
            * ``map``: session-items map,
            * ``time_start``: the earliest event time,
            * ``time_end``: the latest event time,
            * ``longest_items_vector_size``: the number of items in the longest session-sequence,
            * ``number_of_sessions``: total number of parsed sessions,
            * ``metadata``: additional class info.

        """
        pkl = '.pkl'
        if not filename.endswith(pkl):
            filename = filename + pkl

        objects_att_dict = {
            'map': self.session_items_actions_map,
            'time_start': self.time_start,
            'time_end': self.time_end,
            'longest_items_vector_size': self.longest_items_vector_size,
            'number_of_sessions': self.number_of_sessions,
            'metadata': self.metadata
        }

        with open(filename, 'wb') as storage:
            pickle.dump(objects_att_dict, storage)

    def save(self, filename: str):
        """
        Method saves ``Sessions`` class object into a pickled file.

        Parameters
        ----------
        filename : str
            The path to the pickled object. If suffix .pkl is not given then methods appends it into the file.
        """

        pkl = '.pkl'
        if not filename.endswith(pkl):
            filename = filename + pkl

        with open(filename, 'wb') as storage:
            pickle.dump(self, storage)

    def load(self, filename: str):
        """
        Method loads pickled ``Sessions`` class object and assigns its properties and data into the class instance.

        Parameters
        ----------
        filename : str
            The path to the file.

        Raises
        ------
        IOError
            Cannot load data into ``Sessions`` object with parsed events.
        """
        # Check if current object has any items to avoid overwriting
        if self.number_of_sessions > 0:
            raise IOError('Cannot load data into object with sessions! '
                          'Create empty object to load data from pickled file.')

        # Load
        with open(filename, 'rb') as stored_data:
            self.__dict__.update(pickle.load(stored_data).__dict__)

    def __add__(self, other):
        """
        Adds sessions and mapped items from one ``Sessions`` object into the other.

        Parameters
        ----------
        other : Sessions

        Returns
        -------
        merged : Sessions
            The merged and updated ``Sessions`` object.
        """
        merged = Sessions(other.event_session_key,
                          other.event_product_key,
                          other.event_time_key,
                          other.event_action_key)

        merged.session_items_actions_map = merge_dicts(self.session_items_actions_map,
                                                       other.session_items_actions_map)

        merged.time_start = get_smaller_value(self.time_start, other.time_start)
        merged.time_end = get_larger_value(self.time_end, other.time_end)
        merged.longest_items_vector_size = get_larger_value(self.longest_items_vector_size,
                                                            other.longest_items_vector_size)
        merged.number_of_sessions = len(merged.session_items_actions_map)
        return merged

    def __str__(self):
        if self.session_items_actions_map:
            try:
                ts = parse_seconds_to_dt(self.time_start)
                te = parse_seconds_to_dt(self.time_end)
            except ValueError:
                ts = self.time_start
                te = self.time_end
            msg = f'Sessions object statistics:\n' \
                  f'*Number of unique sessions: {self.number_of_sessions}\n' \
                  f'*The longest event stream size per session: {self.longest_items_vector_size}\n' \
                  f'*Period start: {ts}\n' \
                  f'*Period end: {te}'
            return msg
        else:
            msg = 'Empty object. Append events or load pickled session-items map!'
            return msg
